#include "4.5.h"
#include "main_utils.h"
#include <ctime>
#include <unordered_set>
#include <iostream>
#include <string>
#include <vector>
#include <sys/resource.h>

using namespace std;

int GetSize(const CardinalityCounter<string>& c) {
  return c.Size();
}

int GetNumOccurences(const CardinalityCounter<string>& c, const string& s) {
  return c.NumOccurences(s);
}

string GetMostFrequent(const CardinalityCounter<string>& c) {
  return c.MostFrequent();
}

int main() {
  // Limit the memory to 128MB.
  rlimit lim;
  lim.rlim_cur = 128<<20;
  lim.rlim_max = lim.rlim_cur;
  CHECK_EQ(setrlimit(RLIMIT_AS, &lim), 0);
  cerr << "WARNING: This test limits the amount of memory to 128MB.\n"
       << "So if the test fails with an error involving 'std::bad_alloc',"
       << " you're probably using too much memory!\n" << endl;

  // Reproduce the tests of 1.1, and add basic tests for "MostFrequent".
  {
    CardinalityCounter<long long> counter;

    // Basic checks
    CHECK_EQ(0, counter.NumOccurences(123));
    CHECK_EQ(0, counter.NumOccurences(123));
    CHECK_EQ(0, counter.NumOccurences(0));
    CHECK_EQ(0, counter.NumOccurences(-1));
  
    counter.Add(1);
    CHECK_EQ(1, counter.MostFrequent());
    counter.Add(123);
    counter.Add(1234567890123456L);
    counter.Add(123);
    CHECK_EQ(123, counter.MostFrequent());
    counter.Add(123);
    counter.Add(1234567890123456L);
    CHECK_EQ(123, counter.MostFrequent());
  
    CHECK_EQ(1, counter.NumOccurences(1));
    CHECK_EQ(3, counter.NumOccurences(123));
    CHECK_EQ(2, counter.NumOccurences(1234567890123456L));
    CHECK_EQ(1, counter.NumOccurences(1));
    CHECK_EQ(3, counter.NumOccurences(123));
    CHECK_EQ(2, counter.NumOccurences(1234567890123456L));
    CHECK_EQ(0, counter.NumOccurences(-1));
    CHECK_EQ(0, counter.NumOccurences(0));
    CHECK_EQ(3, counter.Size());
  
    counter.Remove(789);
    counter.Remove(789);
    counter.Remove(123);
    counter.Remove(123);
    CHECK_EQ(3, counter.Size());
  
    CHECK_EQ(1, counter.NumOccurences(123));
    CHECK_EQ(1234567890123456L, counter.MostFrequent());
    CHECK_EQ(2, counter.NumOccurences(1234567890123456L));
    CHECK_EQ(1, counter.NumOccurences(1));
    counter.Remove(1234567890123456L);
    counter.Add(1);
    CHECK_EQ(1, counter.MostFrequent());
    counter.Remove(123);
    CHECK_EQ(2, counter.Size());
    counter.Remove(1);
    counter.Remove(1);
    CHECK_EQ(1, counter.Size());
    CHECK_EQ(0, counter.NumOccurences(1));
    counter.Remove(1);
    counter.Remove(1);
    counter.Remove(1);
    counter.Remove(1);
    CHECK_EQ(1, counter.Size());
    CHECK_EQ(0, counter.NumOccurences(1));
    CHECK_EQ(1234567890123456L, counter.MostFrequent());
    counter.Add(1);
    counter.Add(1);
    CHECK_EQ(2, counter.NumOccurences(1));
    CHECK_EQ(1, counter.MostFrequent());
    CHECK_EQ(2, counter.Size());
    counter.Add(123);
    CHECK_EQ(3, counter.Size());
  
    // Also check that cardinalities above 1000000 are supported.
    for (int i = 0; i < 1000001; ++i) counter.Add(456);
    CHECK_EQ(1000001, counter.NumOccurences(456));
    CHECK_EQ(456, counter.MostFrequent());
    CHECK_EQ(4, counter.Size());
    for (int i = 0; i < 1000000; ++i) {
      CHECK_EQ(1000001 - i, counter.NumOccurences(456));
      counter.Remove(456);
    }
    CHECK_EQ(1, counter.NumOccurences(456));
    CHECK_EQ(4, counter.Size());
    counter.Remove(456);
    CHECK_EQ(0, counter.NumOccurences(456));
    CHECK_EQ(3, counter.Size());
    counter.Remove(456);
    CHECK_EQ(0, counter.NumOccurences(456));
    CHECK_EQ(3, counter.Size());
    counter.Add(456);
    CHECK_EQ(4, counter.Size());
  }
  
  // Complexity check: it should all be in constant time.
  {
    // Inserting 100000 elements, 100 times each. Should all take less
    // than a second.
    // 1) Prepare the elements to insert.
    const int kNumElements = 100000;
    const int kCardinality = 100;
    vector<int> elements;
    {
      unordered_set<int> element_set;
      while (element_set.size() < kNumElements) element_set.insert(random());
      elements.insert(elements.begin(), element_set.begin(), element_set.end());
    }
    // 2) Perform the repeated insertions.
    CardinalityCounter<int> counter;
    clock_t c0 = clock();
    for (int i = 0; i < kCardinality; ++i) {
      for (const int e : elements) {
        counter.Add(e); 
      }
    }
    clock_t c1 = clock();
    // 3) Check that all cardinalities are correct.
    for (const int e : elements) CHECK_EQ(kCardinality, counter.NumOccurences(e));
    // 4) Check that the time it took to insert all elements wasn't too long.
    double dt = double(c1 - c0) / CLOCKS_PER_SEC;
    cout << "Checking that your code is fast enough: " << dt << " s" << endl;
    CHECK_LT(dt, 3.0);

    // Now the 1.2 tests.
    CHECK_EQ(kNumElements, counter.Size());
    // Tests the 'const' bit.
    CHECK_EQ(0, GetSize(CardinalityCounter<string>()));
    CHECK_EQ(0, GetNumOccurences(CardinalityCounter<string>(), "Hello"));
  }

  // Check MostFrequent()'s constness, and use strings.
  {
    CardinalityCounter<string> cs;
    cs.Add("Hello");
    CHECK_EQ("Hello", GetMostFrequent(cs));
  }

  // Performance tests for the MostFrequent() call.
  // Scenario 1): alternating bests. This doesn't stress the memory much.
  {
    CardinalityCounter<int> counter;
    const int kRepeat = 500000;
    const clock_t c0 = clock();
    for (int i = 0; i < kRepeat; ++i) {
      counter.Add(123);
      CHECK_EQ(123, counter.MostFrequent());
      counter.Add(456);
      counter.Add(456);
      CHECK_EQ(456, counter.MostFrequent());
      counter.Add(123);
    }
    const clock_t c1 = clock();
    const double dt = double(c1 - c0) / CLOCKS_PER_SEC;
    cout << "Checking that your code is fast enough: " << dt << " s" << endl;
    CHECK_LT(dt, 7.0);
  }

  // Scenario 2): many different elements. Let's say this was already covered
  // in the first performance test above.

  // Performance and correctness test:
  // add/remove elements at random. Remove() should
  // run in amortized constat time, since we have 1/N probability that
  // it picks the most frequent element.
  {
    const int kNumElements = 1000;
    vector<int> elements;
    {
      unordered_set<int> element_set;
      while (element_set.size() < kNumElements) element_set.insert(random());
      elements.insert(elements.begin(), element_set.begin(), element_set.end());
    }
    const int kRepeat = 10000000;
    CardinalityCounter<int> counter;
    const clock_t c0 = clock();
    for (int i = 0; i < kRepeat; ++i) {
      const int r = random();
      const int e = elements[(r >> 1) % elements.size()];
      if (r & 1) counter.Add(e); else counter.Remove(e);
    }
    const clock_t c1 = clock();
    const int size = counter.Size();
    CHECK_GT(size, 3 * kNumElements / 4);
    CHECK_LE(size, kNumElements);
    const int most_frequent = counter.MostFrequent();
    const int occurences = counter.NumOccurences(most_frequent);
    CHECK_LT(occurences, 5 * sqrt(kRepeat / kNumElements));
    CHECK_GT(occurences, sqrt(kRepeat / kNumElements));
    const double dt = double(c1 - c0) / CLOCKS_PER_SEC;
    cout << "Checking that your code is fast enough: " << dt << " s" << endl;
    CHECK_LT(dt, 5.0);
  }

  // Verify that elements with cardinality 0 are really removed:
  // insert 10M elements twice, and remove them twice little after adding them.
  {
    CardinalityCounter<int> counter;
    const clock_t c0 = clock();
    for (int i = 0; i < 10000000; ++i) {
      const int r = random();
      if (r % 2 == 0) {
        const int num = 1 + ((r >> 1) & 4);  // Add it between 1 and 4 times.
        for (int i = 0; i < num; ++i) counter.Add(r);
      } else if (counter.Size() > 0) {
        const int f = counter.MostFrequent();
        const int c = counter.NumOccurences(f);
        for (int i = 0; i < c; ++i) counter.Remove(f);
      }
    }
    const clock_t c1 = clock();
    const double dt = double(c1 - c0) / CLOCKS_PER_SEC;
    cout << "Checking that your code is fast enough: " << dt << " s" << endl;
    CHECK_LT(dt, 10.0);
  }
    
  // Now the fun begins! We'll create some nightmare scenarios that keep changing
  // the max occurence element:
  // 1)
  // - Add a1 1 times, then a2 2 times, then a3 3 times, then a4 4 times, ...
  // - Then remove a_n 2 times, then remove a_n-1 2 times, ...
  // - Then again, and again, and again.
  {
    const int kNumElements = 5000;
    vector<int> elements;
    {
      int seed = 0x47612FED;
      unordered_set<int> element_set;
      while (element_set.size() < kNumElements) {
        element_set.insert(seed *= 1000000007);
      }
      elements.insert(elements.begin(), element_set.begin(), element_set.end());
    }
    CardinalityCounter<int> counter;
    const clock_t c0 = clock();
    int sum = 0;
    for (int i = 0; i < elements.size(); ++i) {
      for (int j = 0; j <= i; ++j) {
        counter.Add(elements[i]);
        sum += counter.MostFrequent();
      }
    }
    while (counter.Size() != 0) {
      for (int i = elements.size() - 1; i >= 0; i--) {
        counter.Remove(elements[i]);
        counter.Remove(elements[i]);
        if (counter.Size() != 0) sum += counter.MostFrequent();
      }
    }
    const clock_t c1 = clock();
    const double dt = double(c1 - c0) / CLOCKS_PER_SEC;
    cout << "Checking that your code is fast enough: " << dt << " s" << endl;
    CHECK_EQ(sum, -754452717);
    CHECK_LT(dt, 6.0);
  }
  // 2)
  // - Add a1, a2, .., an n times on average, at random.
  // - The repeatedly remove the most frequent element a random number of times in
  //   [1..2]. Until it's all empty
  {
    const int kNumElements = 10000;
    const int N = 1000;
    const int MOD = 1000000007;
    vector<int> elements;
    int seed = 0x47612FED;
    {
      unordered_set<int> element_set;
      while (element_set.size() < kNumElements) {
        element_set.insert(seed *= MOD);
      }
      elements.insert(elements.begin(), element_set.begin(), element_set.end());
    }
    CardinalityCounter<int> counter;
    const clock_t c0 = clock();
    int total = kNumElements * N;
    while (--total) counter.Add(elements[(seed *= MOD) % kNumElements]);
    while (counter.Size() != 0) {
      const int f = counter.MostFrequent();
      if ((seed *= MOD) & 1) counter.Remove(f);
      counter.Remove(f);
    }
    const clock_t c1 = clock();
    const double dt = double(c1 - c0) / CLOCKS_PER_SEC;
    cout << "Checking that your code is fast enough: " << dt << " s" << endl;
    CHECK_LT(dt, 7.0);
  }

  cout << "SUCCESS!" << endl;
}
